(*  Title:      HOL/SMT/Tools/smt_link.ML
    Author:     Sascha Boehme, TU Muenchen

FIXME
*)

signature SMT_LINK =
sig
  (*options*)
  val timeout: int Config.T
  val with_timeout: Proof.context -> ('a -> 'b) -> 'a -> 'b
  val trace: bool Config.T
  val trace_msg: Proof.context -> ('a -> string) -> 'a -> unit
  val db_fixed: bool. Config.T
  val set_database: string -> Context.generic -> Context.generic

  (*run solver*)
  datatype solver = Local of string | Remote of string
  val run_solver: solver -> string -> 

  val setup: theory -> theory
end

structure SMT_Link : SMT_DATABASE =
struct

(* hash keys *)

fun hash_key_of filename =
  let
    val hash_cmd = "perl -w " ^ getenv "HASH_KEY_GEN"
    val (out, res) = system_out (hash_cmd ^ " " ^ filename)
  in
    if res <> 0
    then error ("External tool failed to compute hash key: " ^ quote hash_cmd)
    else hd (split_lines out)
  end



(* database *)

type database = {
  db_path: Path.T,
  db_lines: int,
  db_toc: (int * int) Symtab.table}


fun load_db filename =
  let
    val path = Path.explode filename

    fun int_of_string s = fst (read_int (explode s))

    fun err reason = error ("Failed to load database (" ^ reason ^ "): " ^
      quote (Path.implode path))
    fun err_format () = err "(database broken?)" 
    fun err_dup () = err "duplicate key" 

    fun load line ((keys, lc), skip) =
      if skip = 0
      then
        (case space_explode " " line of
          [key, sk] =>
            let val skip' = int_of_string sk
            in
              if Symtab.defined keys key then err_dup ()
              else ((Symtab.update (key, (lc+1, skip')) keys, lc+1), skip')
            end
        | _ => err_format ())
      else ((keys, lc+1), skip-1)

    val (toc, lines) = 
      if File.exists path
      then fst (File.fold_lines load path ((Symtab.empty, 1), 0))
      else (Symtab.empty, 1)
  in {db_path=path, db_lines=lines, db_toc=toc} end


fun get_result db_path (start, len) =
  let
    exception Stop of string list
    fun load line (output, lc) =
      if lc < start then (output, lc+1)
      else if lc < start + len then (line :: output, lc+1)
      else raise Stop (rev output)
  in
    (File.fold_lines load db_path ([], 1);
     error ("Failed to retrieve item from database (database broken?): " ^
       quote (Path.implode db_path)))
    handle Stop output => output
  end


fun add_result key f (io_paths as (_, out_path)) {db_path, db_lines, db_toc} =
  let
    val _ = f io_paths
    val output = rev (the_default [] (try (File.fold_lines cons out_path) []))
    val len = length output
    val _ = NAMED_CRITICAL (Path.implode db_path) (fn () =>
      (File.append db_path (key ^ " " ^ string_of_int len ^ "\n");
       File.append_list db_path output))
    val db = {db_path=db_path, db_lines=db_lines+len+1,
      db_toc=Symtab.update (key, (db_lines+1, len)) db_toc}
  in (output, db) end


fun run_solver f (io_paths as (in_path, _)) (db as {db_path, db_toc, ...}) =
  let val key = hash_key_of (Path.implode in_path)
  in
    (case Symtab.lookup db_toc key of
      SOME loc => (get_result db_path loc, db)
    | NONE => add_result key f io_paths db)
  end




structure Database = Generic_Data
(
  type T = database option
  val empty = NONE
  val extend = I
  fun merge (db1, db2) = if is_some db1 then db1 else db2
)

val (interactive, setup_interactive) = Attrib.config_bool "smt_db_fixed" true

Attrib.setup (Binding.name "smt_db") scan_keep "configuration option"


val setup = setup_interactive #> 



fun with_files ctxt f =
  let
    fun make_names n = (n, n ^ ".proof")

    val keep' = Config.get ctxt keep
    val paths as (problem_path, proof_path) =
      if keep' <> "" andalso File.exists (Path.dir (Path.explode keep'))
      then pairself Path.explode (make_names keep')
      else pairself (File.tmp_path o Path.explode)
        (make_names ("smt-" ^ serial_string ()))

    val y = Exn.capture f (problem_path, proof_path)

    val _ = if keep' = "" then (pairself (try File.rm) paths; ()) else ()
  in Exn.release y end

fun invoke ctxt output f (paths as (problem_path, proof_path)) =
  let
    fun pretty tag ls = Pretty.string_of (Pretty.big_list tag
      (map Pretty.str ls))

    val x = File.open_output output problem_path
    val _ = trace_msg ctxt (pretty "SMT problem:" o split_lines o File.read)
      problem_path

    val _ = with_timeout ctxt f paths
    fun lines_of path = the_default [] (try (File.fold_lines cons path) [])
    val ls = rev (dropwhile (equal "") (lines_of proof_path))
    val _ = trace_msg ctxt (pretty "SMT result:") ls
  in (x, ls) end

val expand_name = Path.implode o Path.expand o Path.explode 

fun run_perl name args ps =
  system_out (space_implode " " ("perl -w" ::
    File.shell_path (Path.explode (getenv name)) ::
    map File.shell_quote args @ ps))

fun use_certificate ctxt ps =
  let val name = Config.get ctxt cert
  in
    if name = "" then false
    else
     (tracing ("Using certificate " ^ quote (expand_name name) ^ " ...");
      run_perl "CERT_SMT_SOLVER" [expand_name name] ps;
      true)
  end

fun run_locally f env_var args ps =
  if getenv env_var = ""
  then f ("Undefined Isabelle environment variable: " ^ quote env_var)
  else
    let val app = Path.expand (Path.explode (getenv env_var))
    in
      if not (File.exists app)
      then f ("No such file: " ^ quote (Path.implode app))
      else
       (tracing ("Invoking local SMT solver " ^ quote (Path.implode app) ^
          " ...");
        system_out (space_implode " " (File.shell_path app ::
        map File.shell_quote args @ ps)); ())
    end

fun run_remote remote_name args ps msg =
  (case remote_name of
    NONE => error msg
  | SOME name =>
      let
        val url = getenv "REMOTE_SMT_URL"
        val _ = tracing ("Invoking remote SMT solver " ^ quote name ^ " at " ^
          quote url ^ " ...")
      in (run_perl "REMOTE_SMT_SOLVER" (url :: name :: args) ps; ()) end)

fun run ctxt {env_var, remote_name} args (problem_path, proof_path) =
  let val ps = [File.shell_path problem_path, ">", File.shell_path proof_path]
  in
    if use_certificate ctxt ps then ()
    else run_locally (run_remote remote_name args ps) env_var args ps
  end

in

fun run_solver ctxt cmd args output =
  with_files ctxt (invoke ctxt output (run ctxt cmd args))



end
